// FIXME
// Copyright 2002 Adam Megacz, see the COPYING file for licensing [GPL]
package org.ibex;

// FEATURE: reflow before allowing js to read from width/height 
// FEATURE: fastpath for rows=1/cols=1
// FEATURE: mark to reflow starting with a certain child
// FEATURE: separate mark_for_reflow and mark_for_resize
// FEATURE: make all methods final
// FEATURE: use a linked list for the "frontier" when packing
// FEATURE:    or else have a way to mark a column "same as last one"?
// FEATURE: reintroduce surface.abort

import java.util.*;
import org.ibex.js.*;
import org.ibex.util.*;
import org.ibex.translators.*;

/**
 *  <p>
 *  Encapsulates the data for a single Ibex box as well as all layout
 *  rendering logic.
 *  </p>
 *
 *  <p>The rendering process consists of four phases; each requires
 *     one DFS pass over the tree</p>
 *  <ol><li> <b>repacking</b>: children of a box are packed into columns
 *           and rows according to their colspan/rowspan attributes and
 *           ordering.
 *  <ol><li> <b>reconstraining</b>: Minimum and maximum sizes of columns are computed.
 *      <li> <b>resizing</b>: width/height and x/y positions of children
 *           are assigned, and PosChange/SizeChanges are triggered.
 *      <li> <b>repainting</b>: children draw their content onto the PixelBuffer.
 *  </ol>
 *
 *  The first three passes together are called the <i>reflow</i> phase.
 *  Reflowing is done in a seperate pass since PosChanges and
 *  SizeChanges trigger an Surface.abort; if rendering were done in the same
 *  pass, rendering work done prior to the Surface.abort would be wasted.
 */
public final class Box extends JSScope implements Scheduler.Task {

    // Macros //////////////////////////////////////////////////////////////////////

    //#define LENGTH int
    //#define MARK_REPACK for(Box b2 = this; b2 != null && !b2.test(REPACK); b2 = b2.parent) b2.set(REPACK);
    //#define MARK_REPACK_b for(Box b2 = b; b2 != null && !b2.test(REPACK); b2 = b2.parent) b2.set(REPACK);
    //#define MARK_REPACK_parent for(Box b2 = parent; b2 != null && !b2.test(REPACK); b2 = b2.parent) b2.set(REPACK);
    //#define MARK_REFLOW for(Box b2 = this; b2 != null && !b2.test(REFLOW); b2 = b2.parent) b2.set(REFLOW);
    //#define MARK_REFLOW_b for(Box b2 = b; b2 != null && !b2.test(REFLOW); b2 = b2.parent) b2.set(REFLOW);
    //#define MARK_RESIZE for(Box b2 = this; b2 != null && !b2.test(RESIZE); b2 = b2.parent) b2.set(RESIZE);
    //#define MARK_RESIZE_b for(Box b2 = b; b2 != null && !b2.test(RESIZE); b2 = b2.parent) b2.set(RESIZE);
    //#define CHECKSET_SHORT(prop) short nu = (short)toInt(value); if (nu == prop) break; prop = nu;
    //#define CHECKSET_INT(prop) int nu = toInt(value); if (nu == prop) break; prop = nu;
    //#define CHECKSET_FLAG(flag) boolean nu = toBoolean(value); if (nu == test(flag)) break; if (nu) set(flag); else clear(flag);
    //#define CHECKSET_BOOLEAN(prop) boolean nu = toBoolean(value); if (nu == prop) break; prop = nu;
    //#define CHECKSET_STRING(prop) if ((value==null&&prop==null)||(value!=null&&value.equals(prop))) break; prop=(String)value;

    void mark_for_repack() { MARK_REPACK; }

    protected Box() { super(null); }

    static Hash boxToCursor = new Hash(500, 3);
    public static final int MAX_LENGTH = Integer.MAX_VALUE;
    static final Font DEFAULT_FONT;
   
    static {
        Font f = null;
        try { f = Font.getFont((Stream)Main.builtin.get("fonts/vera/Vera.ttf"), 10); }
        catch(JSExn e) { Log.info(Box.class, "should never happen: "+e); }
        DEFAULT_FONT = f;
    }

    // FIXME update these
    // box properties can not be trapped
    static final String[] props = new String[] {
        "shrink", "hshrink", "vshrink", "x", "y", "width", "height", "cols", "rows",
        "colspan", "rowspan", "align", "visible", "packed", "globalx", "globaly",
        "minwidth", "maxwidth", "minheight", "maxheight", "indexof", "thisbox", "clip",
        "numchildren", "redirect", "cursor", "mouse"
    };

    // FIXME update these
    // events can have write traps, but not read traps
    static final String[] events = new String[] {
        "Press1", "Press2", "Press3",
        "Release1", "Release2", "Release3",
        "Click1", "Click2", "Click3",
        "DoubleClick1", "DoubleClick2", "DoubleClick3",
        "Enter", "Leave", "Move", 
        "KeyPressed", "KeyReleased", "PosChange", "SizeChange",
        "childadded", "childremoved",
        "Focused", "Maximized", "Minimized", "Close",
        "icon", "titlebar", "toback", "tofront"
    };

    // Flags //////////////////////////////////////////////////////////////////////

    static final int MOUSEINSIDE  = 0x00000001;
    static final int VISIBLE      = 0x00000002;
    static final int PACKED       = 0x00000004;
    static final int HSHRINK      = 0x00000008;
    static final int VSHRINK      = 0x00000010;
    static final int BLACK        = 0x00000020;  // for red-black code

    static final int FIXED        = 0x00000040;
    static final boolean ROWS     = true;
    static final boolean COLS     = false;

    static final int ISROOT       = 0x00000080;
    static final int REPACK       = 0x00000100;
    static final int REFLOW       = 0x00000200;
    static final int RESIZE       = 0x00000400;
    static final int RECONSTRAIN  = 0x00000800;
    static final int ALIGN_TOP    = 0x00001000;
    static final int ALIGN_BOTTOM = 0x00002000;
    static final int ALIGN_LEFT   = 0x00004000;
    static final int ALIGN_RIGHT  = 0x00008000;
    static final int ALIGNS       = 0x0000f000;
    static final int CURSOR       = 0x00010000;  // if true, this box has a cursor in the cursor hash; FEATURE: GC issues?
    static final int CLIP         = 0x00020000;
    static final int STOP_UPWARD_PROPAGATION    = 0x00040000;


    // Instance Data //////////////////////////////////////////////////////////////////////

    Box parent = null;
    Box redirect = this;
    int flags = VISIBLE | PACKED | REPACK | REFLOW | RESIZE | FIXED /* ROWS */ | STOP_UPWARD_PROPAGATION | CLIP;

    private String text = null;
    private Font font = DEFAULT_FONT; 
    private Picture texture = null;
    private short strokewidth = 1;
    public int fillcolor = 0x00000000;
    private int strokecolor = 0xFF000000;

    private int aspect = 0;

    // specified directly by user
    public LENGTH minwidth = 0;
    public LENGTH maxwidth = MAX_LENGTH;
    public LENGTH minheight = 0;
    public LENGTH maxheight = MAX_LENGTH;
    private short rows = 1;
    private short cols = 0;
    private short rowspan = 1;
    private short colspan = 1;

    // computed during reflow
    private short row = 0;
    private short col = 0;
    public LENGTH x = 0;
    public LENGTH y = 0;
    public LENGTH ax = 0;   // FEATURE: roll these into x/y; requires lots of changes
    public LENGTH ay = 0;   // FEATURE: roll these into x/y; requires lots of changes; perhaps y()?
    public LENGTH width = 0;
    public LENGTH height = 0;
    private LENGTH contentwidth = 0;      // == max(minwidth, textwidth, sum(child.contentwidth))
    private LENGTH contentheight = 0;

    /*
    private VectorGraphics.VectorPath path = null;
    private VectorGraphics.Affine transform = null;
    private VectorGraphics.RasterPath rpath = null;
    private VectorGraphics.Affine rtransform = null;
    */

    // Instance Methods /////////////////////////////////////////////////////////////////////

    public final int fontSize() { return font == null ? DEFAULT_FONT.pointsize : font.pointsize; }

    /** invoked when a resource needed to render ourselves finishes loading */
    public void perform() throws JSExn {

        // FIXME; we can't assume that just because we were performed the image is loaded.
        // as external events have occured, check the state of box
        if (texture != null) {
            if (texture.isLoaded) { minwidth = min(texture.width, maxwidth); minheight = min(texture.height, maxheight); }
            else { JS res = texture.stream; texture = null; throw new JSExn("image not found: "+res.unclone()); }
        } else {
            Log.warn(Box.class, "perform() called with null texture");
        }
        MARK_REPACK;
        MARK_REFLOW;
        MARK_RESIZE;
        dirty();
    }

    public Box getRoot() { return parent == null ? this : parent.getRoot(); }
    public Surface getSurface() { return Surface.fromBox(getRoot()); }

    // FEATURE: use cx2/cy2 format
    /** Adds the intersection of (x,y,w,h) and the node's current actual geometry to the Surface's dirty list */
    public void dirty() { dirty(0, 0, width, height); }
    public void dirty(int x, int y, int w, int h) {
        for(Box cur = this; cur != null; cur = cur.parent) {
            // x and y have a different meaning on the root box
            if (cur.parent != null && cur.test(CLIP)) {
                w = min(x + w, cur.width) - max(x, 0);
                h = min(y + h, cur.height) - max(y, 0);
                x = max(x, 0);
                y = max(y, 0);
            }
            if (w <= 0 || h <= 0) return;
            if (cur.parent == null && cur.getSurface() != null) cur.getSurface().dirty(x, y, w, h);
            x += cur.x;
            y += cur.y;
        }
    }


    // Reflow ////////////////////////////////////////////////////////////////////////////////////////

    Box nextPackedSibling() { Box b = nextSibling(); return b == null || (b.test(PACKED | VISIBLE)) ? b : b.nextPackedSibling(); }
    Box firstPackedChild() { Box b = getChild(0); return b == null || (b.test(PACKED | VISIBLE)) ? b : b.nextPackedSibling(); }

    private static Box[] frontier = new Box[65535];
    private static int[] frontier_content = new int[65535];

    /** pack the boxes into rows and columns; also computes contentwidth */
    void repack() {
        for(Box child = getChild(0); child != null; child = child.nextSibling()) child.repack();
        contentwidth = 0;
        contentheight = 0;
        if (treeSize() == 0) { constrain(); return; }
        //#repeat COLS/ROWS rows/cols cols/rows col/row row/col colspan/rowspan rowspan/colspan contentheight/contentwidth contentwidth/contentheight
        if (test(FIXED) == COLS) {
            int childnum = 0;
            Box lastpacked = null;
            int maxfront = 0;
            int maxrow = 0;
            int rowwidth = 0;
            for(Box child = getChild(0); child != null; child = child.nextSibling(), childnum++) {
                if (!(child.test(PACKED) && child.test(VISIBLE))) continue;
                int col = lastpacked == null ? 0 : (lastpacked.col + lastpacked.colspan);
                int row = lastpacked == null ? 0 : lastpacked.row;
                int colspan = min(cols, child.colspan);
                for(int i=0; i<maxfront; i++) {
                    if (col + colspan > cols) {
                        row++;
                        for(; i>0; i--) row = max(row, frontier[i].row + frontier[i].rowspan);
                        col = 0;
                        rowwidth = 0;
                        continue;
                    }
                    Box front = frontier[i];       // FIXME: O(nlgn)
                    if (front.row + front.rowspan <= row) {
                        frontier[i] = frontier[maxfront-1];
                        frontier_content[i] = frontier_content[maxfront-1];
                        maxfront--;
                        frontier[maxfront] = null;
                        frontier_content[maxfront] = 0;
                        i--;
                        continue;
                    }
                    if ((front.col <= col && front.col + front.colspan > col) ||
                        (front.col < (col+colspan) && front.col + front.colspan >= (col+colspan))) {
                        col = front.col + front.colspan;
                        rowwidth += front.contentwidth; // FIXME: suspect
                        i = -1;
                        continue;
                    }
                    break;
                }
                child.col = (short)col;
                child.row = (short)row;
                maxrow = max(maxrow, child.row + child.rowspan);
                contentwidth = max(contentwidth, rowwidth);
                rowwidth = 0;
                lastpacked = child;
                for(int i=0; i<maxfront; i++) {
                    frontier_content[maxfront] =
                        max(frontier_content[maxfront], frontier_content[i] + child.contentheight);
                    contentheight = 
                        max(contentheight, frontier_content[i] + child.contentheight);
                }
                frontier[maxfront++] = child;
            }
            rows = (short)maxrow;
            for(int i=0; i<maxfront; i++) frontier[i] = null;
            for(int i=0; i<maxfront; i++) frontier_content[i] = 0;
        }
        //#end
        constrain();
    }

    void constrain() {
        //#repeat contentwidth/contentheight contentheight/contentwidth minwidth/minheight row/col col/row \
        //        textwidth/textheight maxwidth/maxheight cols/rows rows/cols colspan/rowspan rowspan/colspan
        contentwidth = bound(minwidth,
                             max(contentwidth, font == null || text == null ? 0 : font.textwidth(text)),
                             maxwidth);
        //#end
    }
    
    void resize(LENGTH x, LENGTH y, LENGTH width, LENGTH height) {
        if (x != this.x || y != this.y || width != this.width || height != this.height) {
            boolean sizechange = (this.width != width || this.height != height) && getTrap("SizeChange") != null;
            boolean poschange = (this.x != x || this.y != y) && getTrap("PosChange") != null;
            //do {
                int thisx = parent == null ? 0 : this.x;
                int thisy = parent == null ? 0 : this.y;

                // we can't reenable this until we track
                // surface-relative sizes; imagine the case of a clear
                // surface with nonclear children

                /*
                if (texture == null && (text == null || text.equals(""))) {
                    if ((fillcolor & 0xff000000) == 0) break;
                    // FEATURE: more optimizations here
                    if (this.x == x && this.y == y) {
                        Box who = (parent == null ? this : parent);
                        who.dirty(thisx+min(this.width,width), thisy, Math.abs(width-this.width), max(this.height, height));
                        who.dirty(thisx, thisy+min(this.height,height), min(this.width, width), Math.abs(height-this.height));
                        break;
                    }
                }
                */
                (parent == null ? this : parent).dirty(thisx, thisy, this.width, this.height);
                this.width = width; this.height = height; this.x = x; this.y = y;
                dirty();
                //} while (false);
                //this.width = width; this.height = height; this.x = x; this.y = y;
            if (sizechange) putAndTriggerTrapsAndCatchExceptions("SizeChange", T);
            if (poschange)  putAndTriggerTrapsAndCatchExceptions("PosChange", T);
        }
    }

    private static float[] coeff = null;
    private static LinearProgramming.Simplex lp_h = new LinearProgramming.Simplex(100, 100, 300);
    private static LinearProgramming.Simplex lp = new LinearProgramming.Simplex(100, 100, 300);
    private static int[] regions = new int[65535];
    private static int[] regions_v = new int[65535];

    void place_children() {
        int numkids = 0; for(Box c = firstPackedChild(); c != null; c = c.nextPackedSibling()) numkids++;
        int numregions = 0, numregions_v = 0;
        //#repeat col/row colspan/rowspan contentwidth/contentheight width/height HSHRINK/VSHRINK numregions/numregions_v \
        //        maxwidth/maxheight cols/rows minwidth/minheight lp_h/lp lp_h/lp easy_width/easy_height regions/regions_v
        if (numkids > 0 && cols > 1) do {
            // FIXME: numboxes^2, and damn ugly to boot
            for(Box c = firstPackedChild(); c != null; c = c.nextPackedSibling()) {
                int target = c.col;
                for(boolean stop = false;;) {
                    for(int i=0; i<=numregions; i++) {
                        if (i == numregions) { regions[numregions++] = target; break; }
                        if (target == regions[i]) break;
                        if (target < regions[i]) { int tmp = target; target = regions[i]; regions[i] = tmp; }
                    }
                    if (stop) break;
                    stop = true;
                    target = min(cols, c.col+c.colspan);
                }
            }
            if (regions[numregions-1] == cols) numregions--;
            else regions[numregions] = cols;

            /* boolean easy_width = contentwidth >= width; */
            /*
            for(Box c = firstPackedChild(); easy_width && c != null; c = c.nextPackedSibling()) {
                if (c.contentwidth == c.maxwidth) continue;
                if (c.maxwidth == Integer.MAX_VALUE) continue;
                easy_width = false;
            }
            if (easy_width) for(int i=0; i<cols; i++) {
                easy_width = false;
                boolean good = true;
                for(Box c = firstPackedChild(); good && c != null; c = c.nextPackedSibling())
                    if (c.col <= i && c.col + c.colspan > i && c.maxwidth < Integer.MAX_VALUE)
                        good = false;
                if (good) { easy_width = true; break; }
            }
            if (easy_width) break;
            */
            int nc = numregions * 2 + numkids + 1;
            if (coeff == null || nc+1>coeff.length) coeff = new float[nc+1];
            lp_h.init(nc);

            for(int i=0; i<coeff.length; i++) coeff[i] = (float)0.0;
            coeff[numregions*2+numkids] = (float)10000.0;               // priority 1: sum of columns no greater than parent
            for(int i=numregions*2; i<numregions*2+numkids; i++) coeff[i] = (float)100.0;  // priority 2: honor maxwidths
            for(int i=numregions; i<numregions*2; i++) coeff[i] = (float)(0.1);            // priority 3: equalize columns
            lp_h.setObjective(coeff, false);
            
            for(int i=0; i<numregions; i++) lp_h.set_lowbo(i+1, (float)0.0); // invariant: columns cannot have negative size

            // invariant: columns must be at least as large as parent
            for(int i=0; i<coeff.length; i++) coeff[i] = (i<numregions) ? (float)(regions[i+1] - regions[i]) : (float)0.0;
            lp_h.add_constraint(coeff, LinearProgramming.GE, (float)width);

            // priority 1: sum of columns as close to parent's width as possible
            for(int i=0; i<coeff.length; i++) coeff[i] = (i<numregions) ? (float)(regions[i+1] - regions[i]) : (float)0.0;
            coeff[numregions*2+numkids] = (float)-1.0;
            lp_h.add_constraint(coeff, LinearProgramming.EQ, (float)width);

            int childnum = 0;
            for(Box child = firstPackedChild(); child != null; child = child.nextPackedSibling()) {

                // invariant: honor minwidths
                for(int i=0; i<coeff.length; i++) coeff[i] = (float)0.0;
                for(int r=0; r<numregions; r++)
                    if (regions[r] >= child.col && regions[r+1] <= min(child.col+child.colspan,cols))
                        coeff[r] = (float)(regions[r+1] - regions[r]);
                lp_h.add_constraint(coeff, LinearProgramming.GE, (float)child.contentwidth);

                // priority 2: honor maxwidths
                int child_maxwidth = child.test(HSHRINK) ? min(child.maxwidth, child.contentwidth) : child.maxwidth;
                if (child_maxwidth < Integer.MAX_VALUE) {
                    for(int i=0; i<coeff.length; i++) coeff[i] = (float)0.0;
                    for(int r=0; r<numregions; r++)
                        if (regions[r] >= child.col && regions[r+1] <= min(child.col+child.colspan,cols))
                            coeff[r] = (float)(regions[r+1] - regions[r]);
                    coeff[numregions*2+childnum] = (float)-1.0;
                    lp_h.add_constraint(coeff, LinearProgramming.LE, (float)child_maxwidth);
                }

                childnum++;
            }

            // priority 3: equalize columns
            float avg = ((float)width)/((float)numregions);
            for(int r=0; r<numregions; r++) {
                float weight = (float)(regions[r+1] - regions[r]);
                for(int k=0; k<coeff.length; k++) coeff[k] = (float)(k==r?weight:k==(numregions+r)?-1.0:0.0);
                lp_h.add_constraint(coeff, LinearProgramming.LE, avg * weight);
                for(int k=0; k<coeff.length; k++) coeff[k] = (float)(k==r?weight:k==(numregions+r)?1.0:0.0);
                lp_h.add_constraint(coeff, LinearProgramming.GE, avg * weight);
            }

            try {
                int result = lp_h.solve();
                switch(result) {
                    case LinearProgramming.UNBOUNDED:
                        Log.warn(this, "simplex solver claims unboundedness; this should never happen");
                        break;
                    case LinearProgramming.INFEASIBLE:
                        Log.debug(this, "simplex solver claims infeasibility; this should never happen");
                        break;
                    case LinearProgramming.MILP_FAIL:
                        Log.warn(this, "simplex solver claims MILP_FAIL; this should never happen");
                        break;
                    case LinearProgramming.RUNNING:
                        Log.warn(this, "simplex solver still RUNNING; this should never happen");
                        break;
                    case LinearProgramming.FAILURE:
                        Log.warn(this, "simplex solver claims FAILURE; this should never happen");
                        break;
                }
            } catch (Error e) {
                Log.warn(this, "got an Error in simplex solver; not sure why this happens");
                return;
            }

        } while(false);
        //#end
        
        for(Box child = getChild(0); child != null; child = child.nextSibling()) {
            if (!child.test(VISIBLE)) continue;
            int child_width, child_height, child_x, child_y;
            if (!child.test(PACKED)) {
                child_width = child.test(HSHRINK) ? child.contentwidth : min(child.maxwidth, width - Math.abs(child.ax));
                child_height = child.test(VSHRINK) ? child.contentheight : min(child.maxheight, height - Math.abs(child.ay));
                child_width = max(child.minwidth, child_width);
                child_height = max(child.minheight, child_height);
                int gap_x = width - child_width;
                int gap_y = height - child_height;
                child_x = child.ax + (child.test(ALIGN_RIGHT) ? gap_x : !child.test(ALIGN_LEFT) ? gap_x / 2 : 0);
                child_y = child.ay + (child.test(ALIGN_BOTTOM) ? gap_y : !child.test(ALIGN_TOP) ? gap_y / 2 : 0);
            } else {
                int diff;
                //#repeat col/row colspan/rowspan contentwidth/contentheight width/height colMaxWidth/rowMaxHeight \
                //        child_x/child_y x/y HSHRINK/VSHRINK maxwidth/maxheight cols/rows minwidth/minheight x_slack/y_slack \
                //        child_width/child_height ALIGN_RIGHT/ALIGN_BOTTOM ALIGN_LEFT/ALIGN_TOP lp_h/lp easy_width/easy_height \
                //        numregions/numregions_v regions/regions_v
                child_width = 0;
                child_x = 0;
                if (cols == 1) {
                    child_x = 0;
                    child_width = width;
                    /*
                } else if (easy_width) {
                    */
                } else {
                    for(int r=0; r<numregions; r++) {
                        if (regions[r] >= child.col && regions[r+1] <= min(child.col+child.colspan,cols)) {
                            child_width += Math.round(lp_h.solution[lp_h.rows+r+1] * (regions[r+1] - regions[r]));
                        } else if (regions[r+1] <= child.col) {
                            child_x += Math.round(lp_h.solution[lp_h.rows+r+1] * (regions[r+1] - regions[r]));
                        }
                    }
                }
                diff = (child_width - (child.test(HSHRINK) ? child.contentwidth : min(child_width, child.maxwidth)));
                child_x += (child.test(ALIGN_RIGHT) ? diff : child.test(ALIGN_LEFT) ? 0 : diff / 2);
                child_width -= diff;
                //#end
            }
            child.resize(child_x, child_y, child_width, child_height);
        }

        for(Box child = getChild(0); child != null; child = child.nextSibling())
            if (child.test(VISIBLE) && child.treeSize() > 0)
                child.place_children();
    }



    // Rendering Pipeline /////////////////////////////////////////////////////////////////////

    /** Renders self and children within the specified region. All rendering operations are clipped to xIn,yIn,wIn,hIn */
    void render(int parentx, int parenty, int cx1, int cy1, int cx2, int cy2, PixelBuffer buf, VectorGraphics.Affine a) {
        if (!test(VISIBLE)) return;
        int globalx = parentx + (parent == null ? 0 : x);
        int globaly = parenty + (parent == null ? 0 : y);

        // intersect the x,y,w,h rendering window with ourselves; quit if it's empty
        if (test(CLIP)) {
            cx1 = max(cx1, globalx);
            cy1 = max(cy1, globaly);
            cx2 = min(cx2, globalx + width);
            cy2 = min(cy2, globaly + height);
            if (cx2 <= cx1 || cy2 <= cy1) return;
        }

        if ((fillcolor & 0xFF000000) != 0x00000000 || parent == null)
            buf.fillTrapezoid(cx1, cx2, cy1, cx1, cx2, cy2, (fillcolor & 0xFF000000) == 0 ? 0xffffffff : fillcolor);

        // FIXME: do aspect in here
        if (texture != null && texture.isLoaded)
            for(int x = globalx; x < cx2; x += texture.width)
                for(int y = globaly; y < cy2; y += texture.height)
                    buf.drawPicture(texture, x, y, cx1, cy1, cx2, cy2);
 
        if (text != null && !text.equals("") && font != null) {
            int gap_x = width - font.textwidth(text);
            int gap_y = height - font.textheight(text);
            int text_x = globalx + (test(ALIGN_RIGHT) ? gap_x : !test(ALIGN_LEFT) ? gap_x/2 : 0);
            int text_y = globaly + (test(ALIGN_BOTTOM) ? gap_y : !test(ALIGN_TOP) ? gap_y/2 : 0);
            font.rasterizeGlyphs(text, buf, strokecolor, text_x, text_y, cx1, cy1, cx2, cy2);
        }

        for(Box b = getChild(0); b != null; b = b.nextSibling())
            b.render(globalx, globaly, cx1, cy1, cx2, cy2, buf, null);
    }
    
    
    // Methods to implement org.ibex.js.JS //////////////////////////////////////

    public int globalToLocalX(int x) { return parent == null ? x : parent.globalToLocalX(x - this.x); }
    public int globalToLocalY(int y) { return parent == null ? y : parent.globalToLocalY(y - this.y); }
    public int localToGlobalX(int x) { return parent == null ? x : parent.globalToLocalX(x + this.x); }
    public int localToGlobalY(int y) { return parent == null ? y : parent.globalToLocalY(y + this.y); }
    
    public Object callMethod(Object method, Object a0, Object a1, Object a2, Object[] rest, int nargs) throws JSExn {
        switch (nargs) {
            case 1: {
                //#switch(method)
                case "indexof":
                    Box b = (Box)a0;
                    if (b.parent != this)
                        return (redirect == null || redirect == this) ?
                            N(-1) :
                            redirect.callMethod(method, a0, a1, a2, rest, nargs);
                    return N(b.getIndexInParent());

                case "distanceto":
                    Box b = (Box)a0;
                    JS ret = new JS();
                    ret.put("x", N(b.localToGlobalX(0) - localToGlobalX(0)));
                    ret.put("y", N(b.localToGlobalY(0) - localToGlobalY(0)));
                    return ret;

                //#end
            }
        }
        return super.callMethod(method, a0, a1, a2, rest, nargs);
    }

    public Enumeration keys() { throw new Error("you cannot apply for..in to a " + this.getClass().getName()); }

    protected boolean isTrappable(Object key, boolean isRead) {
        if (key == null) return false;
        else if (key instanceof String) {
            // not allowed to trap box properties, and no read traps on events
            String name = (String)key;
            for (int i=0; i < props.length; i++) if (name.equals(props[i])) return false; 
            if (isRead) for (int i=0; i < events.length; i++) if (name.equals(events[i])) return false; 
        }

        return true;
    }

    public Object get(Object name) throws JSExn {
        if (name instanceof Number)
            return redirect == null ? null : redirect == this ? getChild(toInt(name)) : redirect.get(name);

        //#switch(name)
        case "surface": return parent == null ? null : parent.getAndTriggerTraps("surface");
        case "indexof": return METHOD;
        case "distanceto": return METHOD;
        case "text": return text;
        case "path": throw new JSExn("cannot read from the path property");
        case "fill": return colorToString(fillcolor);
        case "strokecolor": return colorToString(strokecolor);
        case "textcolor": return colorToString(strokecolor);
        case "font": return font == null ? null : font.stream;
        case "fontsize": return font == null ? N(10) : N(font.pointsize);
        case "strokewidth": return N(strokewidth);
        case "align": return alignToString();
        case "thisbox": return this;
        case "shrink": return B(test(HSHRINK) || test(VSHRINK));
        case "hshrink": return B(test(HSHRINK));
        case "vshrink": return B(test(VSHRINK));
        case "aspect": return N(aspect);
        case "x": return (parent == null || !test(VISIBLE)) ? N(0) : N(x);
        case "y": return (parent == null || !test(VISIBLE)) ? N(0) : N(y);
        case "cols": return test(FIXED) == COLS ? N(cols) : N(0);
        case "rows": return test(FIXED) == ROWS ? N(rows) : N(0);
        case "colspan": return N(colspan);
        case "rowspan": return N(rowspan);
        case "width": return N(width);
        case "height": return N(height);
        case "minwidth": return N(minwidth);
        case "maxwidth": return N(maxwidth);
        case "minheight": return N(minheight);
        case "maxheight": return N(maxheight);
        case "clip": return B(test(CLIP));
        case "visible": return B(test(VISIBLE) && (parent == null || (parent.get("visible") == T)));
        case "packed": return B(test(PACKED));
        case "globalx": return N(localToGlobalX(0));
        case "globaly": return N(localToGlobalY(0));
        case "cursor": return test(CURSOR) ? boxToCursor.get(this) : null;
        case "mouse":
            if (getSurface() == null) return null;
            if (getSurface()._mousex == Integer.MAX_VALUE)
                throw new JSExn("you cannot read from the box.mouse property in background thread context");
            return new Mouse();
        case "numchildren": return redirect == null ? N(0) : redirect == this ? N(treeSize()) : redirect.get("numchildren");
        case "redirect": return redirect == null ? null : redirect == this ? T : redirect.get("redirect");
        case "Minimized": if (parent == null && getSurface() != null) return B(getSurface().minimized);
        default: return super.get(name);
        //#end
        throw new Error("unreachable"); // unreachable
    }

    private class Mouse extends JS.Cloneable {
        public Object get(Object key) {
            //#switch(key)
            case "x": return N(globalToLocalX(getSurface()._mousex));
            case "y": return N(globalToLocalY(getSurface()._mousey));

            // this might not get recomputed if we change mousex/mousey...
            case "inside": return B(test(MOUSEINSIDE));
            //#end
            return null;
        }
    }

    void setMaxWidth(Object value) {
        do { CHECKSET_INT(maxwidth); MARK_RESIZE; } while(false);
        if (parent == null && getSurface() != null) getSurface().pendingWidth = maxwidth;
    }
    void setMaxHeight(Object value) {
        do { CHECKSET_INT(maxheight); MARK_RESIZE; } while(false);
        if (parent == null && getSurface() != null) getSurface().pendingHeight = maxheight;
    }

    public void put(Object name, Object value) throws JSExn {
        if (name instanceof Number) { put(toInt(name), value); return; }
        //#switch(name)
        case "text": CHECKSET_STRING(text); MARK_RESIZE; dirty();
        case "strokecolor": value = N(stringToColor((String)value)); CHECKSET_INT(strokecolor); MARK_RESIZE; dirty();
        case "textcolor": value = N(stringToColor((String)value)); CHECKSET_INT(strokecolor); MARK_RESIZE; dirty();
        case "text": CHECKSET_STRING(text); MARK_RESIZE; dirty();
        case "strokewidth": CHECKSET_SHORT(strokewidth); dirty();
        case "shrink": put("hshrink", value); put("vshrink", value);
        case "hshrink": CHECKSET_FLAG(HSHRINK); MARK_RESIZE;
        case "vshrink": CHECKSET_FLAG(VSHRINK); MARK_RESIZE;
        case "width": put("maxwidth", value); put("minwidth", value); MARK_RESIZE;
        case "height": put("maxheight", value); put("minheight", value); MARK_RESIZE;
        case "maxwidth": setMaxWidth(value);
        case "minwidth": CHECKSET_INT(minwidth); MARK_RESIZE;
                         if (parent == null && getSurface() != null)
                             getSurface().setMinimumSize(minwidth, minheight, minwidth != maxwidth || minheight != maxheight);
        case "maxheight": setMaxHeight(value);
        case "minheight": CHECKSET_INT(minheight); MARK_RESIZE;
                         if (parent == null && getSurface() != null)
                             getSurface().setMinimumSize(minwidth, minheight, minwidth != maxwidth || minheight != maxheight);
        case "colspan": if (toInt(value) <= 0) return; CHECKSET_SHORT(colspan); MARK_REPACK_parent;
        case "rowspan": if (toInt(value) <= 0) return; CHECKSET_SHORT(rowspan); MARK_REPACK_parent;
        case "rows": CHECKSET_SHORT(rows); if (rows==0){set(FIXED, COLS);if(cols==0)cols=1;} else set(FIXED, ROWS); MARK_REPACK;
        case "cols": CHECKSET_SHORT(cols); if (cols==0){set(FIXED, ROWS);if(rows==0)rows=1;} else set(FIXED, COLS); MARK_REPACK;
        case "clip": CHECKSET_FLAG(CLIP); if (parent == null) dirty(); else parent.dirty();
        case "visible": CHECKSET_FLAG(VISIBLE); dirty(); MARK_RESIZE; dirty();
        case "packed": CHECKSET_FLAG(PACKED); MARK_REPACK_parent;
        case "aspect": CHECKSET_INT(aspect); dirty();
        case "globalx": put("x", N(globalToLocalX(toInt(value))));
        case "globaly": put("y", N(globalToLocalY(toInt(value))));
        case "align": clear(ALIGNS); setAlign(value == null ? "center" : value); MARK_RESIZE;
        case "cursor": setCursor(value);
        case "fill": setFill(value);
        case "mouse":
            int mousex = toInt(((JS)value).get("x"));
            int mousey = toInt(((JS)value).get("y"));
            getSurface()._mousex = localToGlobalX(mousex);
            getSurface()._mousey = localToGlobalY(mousey);
        case "Minimized": if (parent == null && getSurface() != null) getSurface().minimized = toBoolean(value);  // FEATURE
        case "Maximized": if (parent == null && getSurface() != null) getSurface().maximized = toBoolean(value);  // FEATURE
        case "Close": if (parent == null && getSurface() != null) getSurface().dispose(true);
        case "redirect":
            if (value == null) { redirect = null; return; }
            for(Box cur = (Box)value; cur != null; cur = cur.parent)
                if (cur == redirect) {
                    redirect = (Box)value;
                    return;
                }
            JS.error("redirect can only be set to a descendant of its current value");
        case "font":
            if(!(value instanceof Stream)) throw new JSExn("You can only put streams to the font property");
            font = value == null ? null : Font.getFont((Stream)value, font == null ? 10 : font.pointsize);
            MARK_RESIZE;
            dirty();
        case "fontsize": font = Font.getFont(font == null ? null : font.stream, toInt(value)); MARK_RESIZE; dirty();
        case "x": if (parent==null && Surface.fromBox(this)!=null) {
            CHECKSET_INT(x);
        } else {
            if (test(PACKED) && parent != null) return;
            dirty(); CHECKSET_INT(ax);
            dirty(); MARK_RESIZE;
            dirty();
        }
        case "y": if (parent==null && Surface.fromBox(this)!=null) {
            CHECKSET_INT(y);
        } else {
            if (test(PACKED) && parent != null) return;
            dirty(); CHECKSET_INT(ay);
            dirty(); MARK_RESIZE;
            dirty();
        }
        case "titlebar":
            if (getSurface() != null && value != null) getSurface().setTitleBarText(JS.toString(value));
            super.put(name,value);
            
        case "Press1":        if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "Press2":        if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "Press3":        if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "Release1":      if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "Release2":      if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "Release3":      if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "Click1":        if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "Click2":        if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "Click3":        if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "DoubleClick1":  if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "DoubleClick2":  if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "DoubleClick3":  if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "KeyPressed":    if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "KeyReleased":   if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "Move":          if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);

        case "HScroll":       if (!test(STOP_UPWARD_PROPAGATION) && parent != null)
            parent.putAndTriggerTraps(name, N(((Number)value).floatValue() * ((float)parent.fontSize()) / ((float)fontSize())));
        case "VScroll":       if (!test(STOP_UPWARD_PROPAGATION) && parent != null)
            parent.putAndTriggerTraps(name, N(((Number)value).floatValue() * ((float)parent.fontSize()) / ((float)fontSize())));

        case "_Move":         propagateDownward(name, value, false);
        case "_Press1":       propagateDownward(name, value, false);
        case "_Press2":       propagateDownward(name, value, false);
        case "_Press3":       propagateDownward(name, value, false);
        case "_Release1":     propagateDownward(name, value, false);
        case "_Release2":     propagateDownward(name, value, false);
        case "_Release3":     propagateDownward(name, value, false);
        case "_Click1":       propagateDownward(name, value, false);
        case "_Click2":       propagateDownward(name, value, false);
        case "_Click3":       propagateDownward(name, value, false);
        case "_DoubleClick1": propagateDownward(name, value, false);
        case "_DoubleClick2": propagateDownward(name, value, false);
        case "_DoubleClick3": propagateDownward(name, value, false);
        case "_KeyPressed":   propagateDownward(name, value, false);
        case "_KeyReleased":  propagateDownward(name, value, false);
        case "_HScroll":      propagateDownward(name, value, false);
        case "_VScroll":      propagateDownward(name, value, false);

        case "PosChange":     return;
        case "SizeChange":    return;
        case "childadded":    return;
        case "childremoved":  return;
        case "Enter":         return;
        case "Leave":         return;

        case "thisbox":       if (value == null) removeSelf();

        default:              super.put(name, value);
        //#end
    }

    private String alignToString() {
        switch(flags & ALIGNS) {
            case (ALIGN_TOP | ALIGN_LEFT): return "topleft";
            case (ALIGN_BOTTOM | ALIGN_LEFT): return "bottomleft";
            case (ALIGN_TOP | ALIGN_RIGHT): return "topright";
            case (ALIGN_BOTTOM | ALIGN_RIGHT): return "bottomright";
            case ALIGN_TOP: return "top";
            case ALIGN_BOTTOM: return "bottom";
            case ALIGN_LEFT: return "left";
            case ALIGN_RIGHT: return "right";
            case 0: return "center";
            default: throw new Error("invalid alignment flags: " + (flags & ALIGNS));
        }
    }

    private void setAlign(Object value) {
        //#switch(value)
        case "center": clear(ALIGNS);
        case "topleft": set(ALIGN_TOP | ALIGN_LEFT);
        case "bottomleft": set(ALIGN_BOTTOM | ALIGN_LEFT);
        case "topright": set(ALIGN_TOP | ALIGN_RIGHT);
        case "bottomright": set(ALIGN_BOTTOM | ALIGN_RIGHT);
        case "top": set(ALIGN_TOP);
        case "bottom": set(ALIGN_BOTTOM);
        case "left": set(ALIGN_LEFT);
        case "right": set(ALIGN_RIGHT);
        default: JS.log("invalid alignment \"" + value + "\"");
        //#end
    }
    
    private void setCursor(Object value) {
        if (value == null) { clear(CURSOR); boxToCursor.remove(this); return; }
        if (value.equals(boxToCursor.get(this))) return;
        set(CURSOR);
        boxToCursor.put(this, value);
        Surface surface = getSurface();
        if (surface != null) {
            String tempcursor = surface.cursor;
            propagateDownward(null, null, false);
            if (surface.cursor != tempcursor) surface.syncCursor();
        }
    }

    private void setFill(Object value) throws JSExn {
        if (value == null) {
            // FIXME: Check this... does this make it transparent? 
            texture = null;
            fillcolor = 0;
        } else if (value instanceof String) {
            // FIXME check double set
            int newfillcolor = stringToColor((String)value);
            if (newfillcolor == fillcolor) return;
            fillcolor = newfillcolor;
        } else if(value instanceof JS) {
            texture = Picture.load((JS)value, this);
            if (texture != null && texture.isLoaded) perform();
        } else {
            throw new JSExn("fill must be null, a String, or a stream, not a " + value.getClass());
        }
        dirty();
    }

    // FIXME: mouse move/release still needs to propagate to boxen in which the mouse was pressed and is still held down
    /**
     *  Handles events which propagate down the box tree.  If obscured
     *  is set, then we merely check for Enter/Leave.
     */
    private void propagateDownward(Object name_, Object value, boolean obscured) {

        String name = (String)name_;
        if (getSurface() == null) return;
        int x = globalToLocalX(getSurface()._mousex);
        int y = globalToLocalY(getSurface()._mousey);
        boolean wasinside = test(MOUSEINSIDE);
        boolean isinside = test(VISIBLE) && inside(x, y) && !obscured;
        if (!wasinside && isinside) {
            set(MOUSEINSIDE);
            putAndTriggerTrapsAndCatchExceptions("Enter", T);
        }
        if (isinside && test(CURSOR)) getSurface().cursor = (String)boxToCursor.get(this);
        if (wasinside && !isinside) {
            clear(MOUSEINSIDE);
            putAndTriggerTrapsAndCatchExceptions("Leave", T);
        }

        boolean found = false;
        if (wasinside || isinside)
            for(Box child = getChild(treeSize() - 1); child != null; child = child.prevSibling()) {
                boolean save_stop = child.test(STOP_UPWARD_PROPAGATION);
                Object value2 = value;
                if (name.equals("_HScroll") || name.equals("_VScroll"))
                    value2 = N(((Number)value).floatValue() * ((float)child.fontSize()) / (float)fontSize());
                if (obscured || !child.inside(x - child.x, y - child.y)) {
                    child.propagateDownward(name, value2, true);
                } else try {
                    found = true;
                    child.clear(STOP_UPWARD_PROPAGATION);
                    if (name != null) child.putAndTriggerTrapsAndCatchExceptions(name, value2);
                    else child.propagateDownward(name, value2, obscured);
                } finally {
                    if (save_stop) child.set(STOP_UPWARD_PROPAGATION); else child.clear(STOP_UPWARD_PROPAGATION);
                }
                if (child.inside(x - child.x, y - child.y))
                    if (name != null && name.equals("_Move")) obscured = true;
                    else break;
            }

        if (!obscured && !found)
            if ("_Move".equals(name) || wasinside)
                if (name != null)
                    putAndTriggerTrapsAndCatchExceptions(name.substring(1), value);
    }

    private static int stringToColor(String s) {
        // FIXME support three-char strings by doubling digits
        if (s == null) return 0x00000000;
        else if (SVG.colors.get(s) != null) return 0xFF000000 | toInt(SVG.colors.get(s));
        else if (s.length() == 7 && s.charAt(0) == '#') try {
            // FEATURE  alpha
            return 0xFF000000 |
                (Integer.parseInt(s.substring(1, 3), 16) << 16) |
                (Integer.parseInt(s.substring(3, 5), 16) << 8) |
                Integer.parseInt(s.substring(5, 7), 16);
        } catch (NumberFormatException e) {
            Log.info(Box.class, "invalid color " + s);
            return 0;
        }
        else return 0; // FEATURE: error?
    }

    private static String colorToString(int argb) {
        if ((argb & 0xFF000000) == 0) return null;
        String red = Integer.toHexString((argb & 0x00FF0000) >> 16);
        String green = Integer.toHexString((argb & 0x0000FF00) >> 8);
        String blue = Integer.toHexString(argb & 0x000000FF);
        if (red.length() < 2) red = "0" + red;
        if (blue.length() < 2) blue = "0" + blue;
        if (green.length() < 2) green = "0" + green;
        return "#" + red + green + blue;
    }

    /** figures out what box in this subtree of the Box owns the pixel at x,y relitave to the Surface */
    public static Box whoIs(Box cur, int x, int y) {

        if (cur.parent != null) throw new Error("whoIs may only be invoked on the root box of a surface");
        int globalx = 0;
        int globaly = 0;

        // WARNING: this method is called from the event-queueing thread -- it may run concurrently with
        // ANY part of Ibex, and is UNSYNCHRONIZED for performance reasons.  BE CAREFUL HERE.

        if (!cur.test(VISIBLE)) return null;
        if (!cur.inside(x - globalx, y - globaly)) return cur.parent == null ? cur : null;
        OUTER: while(true) {
            for(int i=cur.treeSize() - 1; i>=0; i--) {
                Box child = cur.getChild(i);
                if (child == null) continue;        // since this method is unsynchronized, we have to double-check
                globalx += child.x;
                globaly += child.y;
                if (child.test(VISIBLE) && child.inside(x - globalx, y - globaly)) { cur = child; continue OUTER; }
                globalx -= child.x;
                globaly -= child.y;
            }
            break;
        }
        return cur;
    }


    // Trivial Helper Methods (should be inlined) /////////////////////////////////////////

    static short min(short a, short b) { if (a<b) return a; else return b; }
    static int min(int a, int b) { if (a<b) return a; else return b; }
    static float min(float a, float b) { if (a<b) return a; else return b; }

    static short max(short a, short b) { if (a>b) return a; else return b; }
    static int max(int a, int b) { if (a>b) return a; else return b; }
    static float max(float a, float b) { if (a>b) return a; else return b; }

    static int min(int a, int b, int c) { if (a<=b && a<=c) return a; else if (b<=c && b<=a) return b; else return c; }
    static int max(int a, int b, int c) { if (a>=b && a>=c) return a; else if (b>=c && b>=a) return b; else return c; }
    static int bound(int a, int b, int c) { if (c < b) return c; if (a > b) return a; return b; }
    final boolean inside(int x, int y) { return test(VISIBLE) && x >= 0 && y >= 0 && x < width && y < height; }

    void set(int mask) { flags |= mask; }
    void set(int mask, boolean setclear) { if (setclear) set(mask); else clear(mask); }
    void clear(int mask) { flags &= ~mask; }
    boolean test(int mask) { return ((flags & mask) == mask); }
    

    // Tree Handling //////////////////////////////////////////////////////////////////////

    public final int getIndexInParent() { return parent == null ? 0 : parent.indexNode(this); }
    public final Box nextSibling() { return parent == null ? null : parent.getChild(parent.indexNode(this) + 1); }
    public final Box prevSibling() { return parent == null ? null : parent.getChild(parent.indexNode(this) - 1); }
    public final Box getChild(int i) {
        if (i < 0) return null;
        if (i >= treeSize()) return null;
        return (Box)getNode(i);
    }

    // Tree Manipulation /////////////////////////////////////////////////////////////////////

    void removeSelf() {
        if (parent != null) { parent.removeChild(parent.indexNode(this)); return; }
        Surface surface = Surface.fromBox(this); 
        if (surface != null) surface.dispose(true);
    }

    /** remove the i^th child */
    public void removeChild(int i) {
        Box b = getChild(i);
        MARK_REFLOW_b;
        b.dirty();
        b.clear(MOUSEINSIDE);
        deleteNode(i);
        b.parent = null;
        MARK_REFLOW;
        putAndTriggerTrapsAndCatchExceptions("childremoved", b);
    }
    
    public void put(int i, Object value) throws JSExn {
        if (i < 0) return;
            
        if (value != null && !(value instanceof Box)) {
            if (Log.on) JS.warn("attempt to set a numerical property on a box to a non-box");
            return;
        }

        if (redirect == null) {
            if (value == null) putAndTriggerTrapsAndCatchExceptions("childremoved", getChild(i));
            else JS.warn("attempt to add/remove children to/from a node with a null redirect");

        } else if (redirect != this) {
            if (value != null) putAndTriggerTrapsAndCatchExceptions("childadded", value);
            redirect.put(i, value);
            if (value == null) {
                Box b = (Box)redirect.get(new Integer(i));
                if (b != null) putAndTriggerTrapsAndCatchExceptions("childremoved", b);
            }

        } else if (value == null) {
            if (i < 0 || i > treeSize()) return;
            Box b = getChild(i);
            removeChild(i);
            putAndTriggerTrapsAndCatchExceptions("childremoved", b);

        } else {
            Box b = (Box)value;

            // check if box being moved is currently target of a redirect
            for(Box cur = b.parent; cur != null; cur = cur.parent)
                if (cur.redirect == b) {
                    if (Log.on) JS.warn("attempt to move a box that is the target of a redirect");
                    return;
                }

            // check for recursive ancestor violation
            for(Box cur = this; cur != null; cur = cur.parent)
                if (cur == b) {
                    if (Log.on) JS.warn("attempt to make a node a parent of its own ancestor");
                    if (Log.on) Log.info(this, "box == " + this + "  ancestor == " + b);
                    return;
                }

            if (b.parent != null) b.parent.removeChild(b.parent.indexNode(b));
            insertNode(i, b);
            b.parent = this;
            
            // need both of these in case child was already uncalc'ed
            MARK_REFLOW_b;
            MARK_REFLOW;
            
            b.dirty(); 
            putAndTriggerTrapsAndCatchExceptions("childadded", b);
        }
    }

    void putAndTriggerTrapsAndCatchExceptions(Object name, Object val) {
        try {
            putAndTriggerTraps(name, val);
        } catch (JSExn e) {
            JS.log("caught js exception while putting to trap \""+name+"\"");
            JS.log(e);
        } catch (Exception e) {
            JS.log("caught exception while putting to trap \""+name+"\"");
            JS.log(e);
        }
    }

}






        /*
        offset_x = 0;
        if (path != null) {
            if (rpath == null) rpath = path.realize(transform == null ? VectorGraphics.Affine.identity() : transform);
            if ((flags & HSHRINK) != 0) contentwidth = max(contentwidth, rpath.boundingBoxWidth());
            if ((flags & VSHRINK) != 0) contentheight = max(contentheight, rpath.boundingBoxHeight());
            // FIXME: separate offset_x needed for the path
        }
        // #repeat x1/y1 x2/y2 x3/y3 x4/y4 contentwidth/contentheight left/top right/bottom
        int x1 = transform == null ? 0 : (int)transform.multiply_px(0, 0);
        int x2 = transform == null ? 0 : (int)transform.multiply_px(contentwidth, 0);
        int x3 = transform == null ? contentwidth : (int)transform.multiply_px(contentwidth, contentheight);
        int x4 = transform == null ? contentwidth : (int)transform.multiply_px(0, contentheight);
        int left = min(min(x1, x2), min(x3, x4));
        int right = max(max(x1, x2), max(x3, x4));
        contentwidth = max(contentwidth, right - left);
        offset_x = -1 * left;
        // #end
        */


                    /*
        if (path != null) {
            if (rtransform == null) rpath = null;
            else if (!rtransform.equalsIgnoringTranslation(a)) rpath = null;
            else {
                rpath.translate((int)(a.e - rtransform.e), (int)(a.f - rtransform.f));
                rtransform = a.copy();
            }
            if (rpath == null) rpath = path.realize((rtransform = a) == null ? VectorGraphics.Affine.identity() : a);
            if ((strokecolor & 0xff000000) != 0) rpath.stroke(buf, 1, strokecolor);
            if ((fillcolor & 0xff000000) != 0) rpath.fill(buf, new VectorGraphics.SingleColorPaint(fillcolor));
        }
*/


/*
            VectorGraphics.Affine a2 = VectorGraphics.Affine.translate(b.x, b.y);
            if (transform != null) a2.multiply(transform);
            a2.multiply(VectorGraphics.Affine.translate(offset_x, offset_y));
            a2.multiply(a);
*/
